// Copyright (C) 2016 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only

#pragma once

#include <QtGui/QPalette>
#include <QtCore/QTextStream>
#include <QtGui/private/qmath_p.h>
#include <QtGui/private/qtextengine_p.h>

#include <QtGui/private/qt_egl_p.h>

QT_BEGIN_NAMESPACE

QT_END_NAMESPACE
